import numpy as np
from timeit import default_timer as timer
import ctypes
import os
import fnmatch

from .phot import flux, flux_ng
from .kep import Kepler
from .orbits import *
__all__ = ['HierarchicalSystem', 'ConfocalSystem']

path, _ = os.path.split(__file__)
libname = fnmatch.filter(os.listdir(path), 'conf*.so')[0]
conflib = ctypes.CDLL(path + "/" + libname)
libname = fnmatch.filter(os.listdir(path), 'hrch*.so')[0]
hrchlib = ctypes.CDLL(path + "/" + libname)

confargs = [
    'a1', 
    't1', 
    'e1', 
    'p1', 
    'w1', 
    'i1',      
    'a2', 
    't2', 
    'e2', 
    'p2', 
    'o2', 
    'w2', 
    'i2', 
]

hrchargs = [
    'a1', 
    't1', 
    'e1', 
    'p1', 
    'w1', 
    'i1',      
    'a2', 
    't2', 
    'e2', 
    'p2', 
    'o2', 
    'w2', 
    'i2', 
    'm2'
]

class System:
    
    """Base class of all systems. This class should not 
    be instantiated directly. Use ConfocalSystem or 
    HierarchicalSystem instead.
    
    Args:
        o1 (Orbit): The orbit of the first body 
        o2 (Orbit): The orbit of the second body 
    """
    
    def __init__(self, o1, o2):
        
        self.o1 = o1
        self.o2 = o2
        self.pdict = {**o1.pdict(), **o2.pdict()}
            
    def coords(self, t):
        
        """
        Get the coordinates of the two bodies.
        
        Args:
            t (array): Times at which the coordinates should be computed.
            
        Returns:
            tuple: The cartesian coordinates of each body.
        """
        
        x1, y1, z1, x2, y2, z2 = self.kep.coords(t, self.pdict)
        return ((x1, y1, z1), 
                (x2, y2, z2))
    
    def impacts(self, t, grad=False):
        
        if grad:
            bp, bpm, theta, dbp, dbpm, dtheta = self.kep.grad_impacts(
                t, 
                self.pdict
            )
            return bp, bpm, theta, dbp, dbpm, dtheta
        else:
            bp, bpm, theta = self.kep.impacts(t, self.pdict)
            return bp, bpm, theta
        
    def phot(self, t, u1, u2, r1, r2, bp, bpm, theta, grad=False):
        
        if grad:
            lc = flux(
                u1, 
                u2, 
                r1, 
                r2, 
                bp, 
                bpm, 
                np.cos(theta), 
                np.sin(theta)
            ).T
        else:
            lc = flux_ng(
                u1, 
                u2, 
                r1, 
                r2, 
                bp, 
                bpm, 
                np.cos(theta), 
                np.sin(theta)
            )
        
        return lc
    
    def lightcurve(self, t, u1, u2, r1, r2, grad=False, integrate=None, dt=None):
        
        """
        Get the lightcurve resulting from a transit of the moon/planet system.
        
        Args: 
            t (array): Times at which the flux should be computed
            r1 (float): Radius of the body in the PrimaryOrbit
            r2 (float): Radius of the body in the SecondaryOrbit or ConfocalOrbit
            u1 (float): The first limb-darkening parameter
            u2 (float): The second limb-darkening parameter
            grad (bool): If True, compute the gradient of the lightcurve.
                Default is False. 
                
        Returns:
            array: The lightcurve and, if grad=True, a dictionary containing the derivatives with respect to each parameter.
            
        """
        if integrate == 'simpson':
            ta = t - dt / 2
            tb = t + dt / 2
            
            if grad:
                f, g = self.lightcurve(t, u1, u2, r1, r2, grad=True)
                fa, ga = self.lightcurve(ta, u1, u2, r1, r2, grad=True)
                fb, gb = self.lightcurve(tb, u1, u2, r1, r2, grad=True)
                lc = (fa + 4 * f + fb) / 6
                grad = {
                        ka: (va + 4 * v + vb) / 6
                        for (k, v), (ka, va), (kb, vb) in zip(g.items(), ga.items(), gb.items())
                }
                return lc, grad
            else:
                f = self.lightcurve(t, u1, u2, r1, r2, grad=False)
                fa = self.lightcurve(ta, u1, u2, r1, r2, grad=False)
                fb = self.lightcurve(tb, u1, u2, r1, r2, grad=False)
                return (fa + 4 * f + fb) / 6
        
        if integrate == 'trapezoid':
            
            if dt is None:
                dt = np.diff(t)
                tt = np.zeros(len(t) + 1)
                tt[1:-1] = t[1:] - dt / 2
                tt[0] = t[0] - dt[0] / 2
                tt[-1] = t[-1] + dt[-1] / 2
                
                if grad:
                    f, g = self.lightcurve(tt, u1, u2, r1, r2, grad=True)
                    lc = (f[1:] + f[:-1]) / 2
                    grad = {
                        k: (v[1:] + v[:-1]) / 2
                        for k, v in g.items()
                    }
                    return lc, grad
                else:
                    f = self.lightcurve(tt, u1, u2, r1, r2, grad=False)
                    return (f[1:] + f[:-1]) / 2
            else:
            
                ta = t - dt / 2
                tb = t + dt / 2
                if grad:
                    fa, ga = self.lightcurve(ta, u1, u2, r1, r2, grad=True)
                    fb, gb = self.lightcurve(tb, u1, u2, r1, r2, grad=True)
                    lc = (fa + fb) / 2
                    grad = {
                        ka: (va + vb) / 2
                        for (ka, va), (kb, vb) in zip(ga.items(), gb.items())
                    }
                    return lc, grad
                else:
                    fa = self.lightcurve(ta, u1, u2, r1, r2)
                    fb = self.lightcurve(tb, u1, u2, r1, r2)
                    return (fa + fb) / 2
        
        if grad:
            bp, bpm, theta, dbp, dbpm, dtheta = self.kep.grad_impacts(
                t, 
                self.pdict
            )
            f = flux(
                u1, 
                u2, 
                r1, 
                r2, 
                bp, 
                bpm, 
                np.cos(theta), 
                np.sin(theta)
            ).T
                
            lc = f[0]
            f_bp = f[3]
            f_bpm = f[4]
            f_theta = f[5]
                
            df = (
                f_bp * dbp 
                + f_bpm * dbpm 
                + f_theta * dtheta
            )
                
            grad = {
                self.argnames[i]: df[i] 
                for i in range(np.shape(df)[0])
            }
            
            # order?
            grad['r1'] = f[1]
            grad['r2'] = f[2]
            grad['u1'] = f[6]
            grad['u2'] = f[7]
            
            return lc, grad
        
        else:
            bp, bpm, theta = self.kep.impacts(
                t, 
                self.pdict
            )
            return flux_ng(
                u1, 
                u2, 
                r1, 
                r2, 
                bp, 
                bpm, 
                np.cos(theta), 
                np.sin(theta)
            )
                    
    def loglike(self, y, t, u1, u2, r1, r2, sigma, integrate=None, dt=None, grad=False):
        
        """
        Get the log-likelihood of the lightcurve with respect to a set of observations.
        
        Args:
            y (array): A vector of observations to compute the likelihood with 
                respect to. 
            t (array): Times at which the flux should be computed
            r1 (float): Radius of the body in the PrimaryOrbit
            r2 (float): Radius of the body in the SecondaryOrbit or ConfocalOrbit
            u1 (float): The first limb-darkening parameter
            u2 (float): The second limb-darkening parameter
            sigma (float): The standard deviation of the model
            
        Returns: 
            float: The natural logarithm of the likelihood
        """
        
        if grad:
            mu, jac = self.lightcurve(t, u1, u2, r1, r2, integrate=integrate, dt=dt, grad=True)
            s2 = sigma * sigma
            jac = np.array(list(jac.values()))
            ll = -0.5 * np.sum((y - mu) ** 2 / s2 + np.log(s2))
            dldsig = np.sum((y - mu) ** 2 / (s2 * sigma) - 1 / sigma)
            dldx = np.sum((y - mu) * jac / s2, axis=1)
            return ll, np.hstack([dldsig, dldx])
        else:
            mu = self.lightcurve(t, u1, u2, r1, r2, integrate=None, dt=None)
            s2 = sigma * sigma
            return -0.5 * np.sum((y - mu) ** 2 / s2 + np.log(s2))

    def time(self, t, u1, u2, r1, r2, phot_only=False, grad=False, ntimes=1):
        
        if phot_only:
            bp, bpm, theta = self.kep.impacts(t)
            if grad:
                start = timer()
                for _ in range(ntimes):
                    lc = self.phot(t, u1, u2, r1, r2, bp, bpm, theta, grad=True)
                end = timer()
            else:
                start = timer()
                for _ in range(ntimes):
                    lc = self.phot(t, u1, u2, r1, r2, bp, bpm, theta)
                end = timer()
        else:
            if grad:
                start = timer()
                for _ in range(ntimes):
                    bp, bpm, theta, dbp, dbpm, dtheta = self.kep.impacts(t, grad=True)
                    lc = self.phot(t, u1, u2, r1, r2, bp, bpm, theta, grad=True)
                    df = (
                        lc[3] * dbp 
                        + lc[4] * dbpm 
                        + lc[5] * dtheta
                    )
                end = timer()
            else:
                start = timer()
                for _ in range(ntimes):
                    bp, bpm, theta = self.kep.impacts(t)
                    lc = self.phot(t, u1, u2, r1, r2, bp, bpm, theta)
                end = timer()
        return (end - start) / ntimes
    
class ConfocalSystem(System):
    
    """
    Represents a system with two bodies 
    orbiting a central star. 
    
    Args:
        o1 (PrimaryOrbit): The orbit of the first body 
        o2 (ConfocalOrbit): The orbit of the second body
    """
    
    def __init__(self, o1, o2):
        super().__init__(o1, o2)
        
        if not isinstance(o1, PrimaryOrbit):
            msg = ("o1 should be a PrimaryOrbit")
            raise AttributeError(msg)
        if not isinstance(o2, ConfocalOrbit):
            msg = ("o2 should be a ConfocalOrbit")
            raise AttributeError(msg)
        
        self.argnames = confargs
        self.kep = Kepler(self.argnames, conflib)
        
class HierarchicalSystem(System):
    
    """
    Represents a system with a primary body orbiting 
    a central star and a satellite orbiting the 
    primary body.
    
    Args:
        o1 (PrimaryOrbit): The orbit of the primary 
                           body around the central star
        o2 (SatelliteOrbit): The orbit of the satellite around 
                             the primary body.
    """
    
    def __init__(self, o1, o2):
        super().__init__(o1, o2)
        
        if not isinstance(o1, PrimaryOrbit):
            msg = ("o1 should be a PrimaryOrbit")
            raise AttributeError(msg)
        if not isinstance(o2, SatelliteOrbit):
            msg = ("o2 should be a SatelliteOrbit")
            raise AttributeError(msg)
        
        self.argnames = hrchargs
        self.kep = Kepler(self.argnames, hrchlib) 
        




import pytest
import numpy as np

import utils as tu
import gefera as gf

flux = gf.phot.flux
flux_ng = gf.phot.flux_ng

coords = [
        [0.95, 1, 1.15, 0.75],
        [0.95, 1, 0.7, 0.75],
        [0.8, 0.8, 1.15, 0.75],
        [0.4, 0.4, 0.4, 0.4],
        [0.4, 0.4, 0.6, 0.2],
        [0.4, 0.4, 0.65, 0.65],
        [0.55, 0.55, 0.52, 0.52],
        [0.6, 0.6, 0.4, 0.4],
        [0.65, 0.65, 0.65, 0.65],
        [0.6, 0.6, 0.75, 0.75],
        [0.8, 0.8, 0.65, 0.65],
        [0.52, 0.52, 0.7, 0.7],
        [0.88, 0.88, 0.7, 0.7],
        [0.85, 0.9, 1.05, 0.5],
        [0.5, 0.55, 0.8, 0.3],
        [0.6, 0.65, 0.9, 0.4]
    ]

def test_flux():
    
    q1, q2 = np.random.rand(2)
    u1 = 2 * np.sqrt(q1) * q2
    u2 = np.sqrt(q1) * (1 - 2 * q2)
    
    rp, rm = 0.3, 0.2
    
    for xp, yp, xm, ym in coords:
        bp, bpm, theta = tu.change_coords(xp, yp, xm, ym)    
    
        analytic = flux(
            u1, u2, rp, rm, 
            np.array([bp]), 
            np.array([bpm]), 
            np.array([np.cos(theta)]), 
            np.array([np.sin(theta)])
        )[0][0]
    
        numerical = tu.flux_numerical(
            u1, u2, rp, rm, bp, bpm, theta, 500000
        )
        assert np.isclose(numerical, analytic, atol=1e-4)
        
def test_flux_ng():
    
    q1, q2 = np.random.rand(2)
    u1 = 2 * np.sqrt(q1) * q2
    u2 = np.sqrt(q1) * (1 - 2 * q2)
    
    rp, rm = 0.3, 0.2

    for xp, yp, xm, ym in coords:
        bp, bpm, theta = tu.change_coords(xp, yp, xm, ym)    
    
        analytic = flux_ng(
            u1, u2, rp, rm, 
            np.array([bp]), 
            np.array([bpm]), 
            np.array([np.cos(theta)]), 
            np.array([np.sin(theta)])
        )[0]
    
        numerical = tu.flux_numerical(
            u1, u2, rp, rm, bp, bpm, theta, 500000
        )
        assert np.isclose(numerical, analytic, atol=1e-4)
        
def test_flux_grad():
    
    u1, u2, rp, rm = 0.5, 0.3, 0.1, 0.05
    params = [u1, u2, rp, rm]
    dx = 1e-8
    
    for xp, yp, xm, ym in coords:
        bp, bpm, theta = tu.change_coords(xp, yp, xm, ym)    
    
        lc = flux(
            params[0], params[1], params[2], params[3], 
            np.array([bp]), 
            np.array([bpm]), 
            np.array([np.cos(theta)]), 
            np.array([np.sin(theta)])
        )[0]
        exact = [lc[6], lc[7], lc[1], lc[2]]
    
        fd = np.zeros_like(params)
        for i in range(len(params)):
            
            params[i] += dx
            
            flux_plus = flux_ng(
                params[0], params[1], params[2], params[3], 
                np.array([bp]), 
                np.array([bpm]), 
                np.array([np.cos(theta)]), 
                np.array([np.sin(theta)])
            )[0]
            
            params[i] -= 2 * dx
            flux_minus = flux_ng(
                params[0], params[1], params[2], params[3], 
                np.array([bp]), 
                np.array([bpm]), 
                np.array([np.cos(theta)]), 
                np.array([np.sin(theta)])
            )[0]
        
            fd[i] = (flux_plus - flux_minus) / (2 * dx)
    
        assert np.allclose(fd, exact, rtol=1e-5)


import numpy as np
from timeit import default_timer as timer
import ctypes
import os
import fnmatch

__all__ = ['PrimaryOrbit', 'SatelliteOrbit', 'ConfocalOrbit']

class Orbit:
    
    """Parent class of all orbits. This class does not contain
    all the necessary attributes for all orbits. 
    
    Args:
        a: Semimajor axis
        t: Time of periastron passage
        e: Eccentricity
        p: Period
        w: Argument of periastron (in radians)
        i: Inclination (in radians)
    """
    
    def __init__(self, a, t, e, p, w, i):
        
        self.a = a
        self.t = t
        self.e = e
        self.p = p
        self.w = w
        self.i = i
        
    def pdict(self):
        
        return vars(self)
    
class PrimaryOrbit(Orbit):
    
    """
    A heliocentric orbit for the primary body in the system. 
    
    Args:
        a: Semimajor axis
        t: Time of periastron passage
        e: Eccentricity
        p: Period
        w: Argument of periastron (in radians)
        i: Inclination (in radians)
    """
    
    def __init__(self, a, t, e, p, w, i):
        super().__init__(a, t, e, p, w, i)
        
    def pdict(self):
        
        return {k + '1': v for k, v in vars(self).items()}
    
class SatelliteOrbit(Orbit):
    
    """
    The orbit of the moon around the planet.
    
    Args:
        a: Semimajor axis
        t: Time of periastron passage
        e: Eccentricity
        p: Period
        o: Longitude of ascending node (in radians)
        w: Argument of periastron (in radians)
        i: Inclination (in radians)
        m: Moon/planet mass ratio
    """
    
    def __init__(self, a, t, e, p, o, w, i, m):
        super().__init__(a, t, e, p, w, i)
        self.o = o
        self.m = m
        
    def pdict(self):
        
        return {k + '2': v for k, v in vars(self).items()}
        
class ConfocalOrbit(Orbit):
    
    """
    A second heliocentric orbit. 
    
    Args:
        a: Semimajor axis
        t: Time of periastron passage
        e: Eccentricity
        p: Period
        o: Longitude of ascending node (in radians)
        w: Argument of periastron (in radians)
        i: Inclination (in radians)
    """
    
    def __init__(self, a, t, e, p, o, w, i):
        super().__init__(a, t, e, p, w, i)
        self.o = o
        
    def pdict(self):
        
        return {k + '2': v for k, v in vars(self).items()}


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
import gefera as gf

def draw(sys, ax, t, r1, r2, ld_params=None, cmap=plt.cm.copper, fill=True):
    
    """
    Draw the state of the system.
    
    Args:
        sys (System): The system
        ax (matplotlib.pyplot.axis): Axis on which to draw
        t (float): Time at which to draw snapshot of system
        r1 (float): Radius of the first body
        r2 (float): Radius of the second body
        ld_params (optional, float): Limb-darkening parameters for the star
        cmap (optional, matplotlib.colors.Colormap): Colormap for coloring the star according to ld_params
        fill (optional, bool): If true, fill transiting bodies, if false draw outlines only
    """
        
    if isinstance(t, np.ndarray):
        raise Exception("Argument t should be a scalar, not an array.")
            
    x1, y1, _, x2, y2, _ = sys.kep.coords(np.array([t]), sys.pdict) 
        
    b1 = plt.Circle(
        (-x1, y1),
        radius=r1,
        color='k',
        fill=fill
    )
    b2 = plt.Circle(
        (-x2, y2),
        radius=r2,
        color='k',
        fill=fill
    )
    if ld_params is None:
        star = plt.Circle((0, 0), radius=1, color=cmap(1.0), fill=True)
        ax.add_patch(star)
        
    else:
        u1, u2 = ld_params
            
        d = 0.01
        x = np.arange(-1.0, 1.0, d)
        y = np.arange(-1.0, 1.0, d)
        x, y = np.meshgrid(x, y)
        shape = np.shape(x)
        x = x.flatten()
        y = y.flatten()
        z = np.zeros(len(x))
        with np.errstate(invalid='ignore'):
            for i, (x, y) in enumerate(zip(x, y)):
                z[i] = 1 - u1 * (1 - np.sqrt(1 - x**2 - y**2)) - u2 * (1 - np.sqrt(1 - x**2 - y**2))**2
        z = z.reshape(shape)
            
        cmap = cmap
        cmap.set_bad('white')

        edge = plt.Circle(
            (0.005, -0.0), 
            radius=1.0, 
            color='w',
            fill=False, 
            linewidth=4
        )
            
        ax.imshow(z, interpolation='bilinear', extent=(-1, 1, -1, 1), cmap=cmap)
        ax.add_patch(edge)
        
    ax.add_patch(b1)
    ax.add_patch(b2)
    ax.set_xlim(-1.5, 1.5)
    ax.set_ylim(-1.5, 1.5)
    ax.set_axis_off()
    
def draw_series(sys, ax, t, r1, r2, ld_params=None, cmap=plt.cm.copper, fill=True, alpha=1.0):
            
    if ld_params is None:
        star = plt.Circle((0, 0), radius=1, color=cmap(1.0), fill=True)
        ax.add_patch(star)
        
    else:
        u1, u2 = ld_params
            
        d = 0.01
        x = np.arange(-1.0, 1.0, d)
        y = np.arange(-1.0, 1.0, d)
        x, y = np.meshgrid(x, y)
        shape = np.shape(x)
        x = x.flatten()
        y = y.flatten()
        z = np.zeros(len(x))
        with np.errstate(invalid='ignore'):
            for i, (x, y) in enumerate(zip(x, y)):
                z[i] = 1 - u1 * (1 - np.sqrt(1 - x**2 - y**2)) - u2 * (1 - np.sqrt(1 - x**2 - y**2))**2
        z = z.reshape(shape)
            
        cmap = cmap
        cmap.set_bad('white')

        edge = plt.Circle(
            (0.005, -0.0), 
            radius=1.0, 
            color='w',
            fill=False, 
            linewidth=4
        )
            
        ax.imshow(z, interpolation='bilinear', extent=(-1, 1, -1, 1), cmap=cmap)
        ax.add_patch(edge)
    
    for ti in t:
        x1, y1, _, x2, y2, _ = sys.kep.coords(np.array([ti]), sys.pdict) 
        
        b1 = plt.Circle(
            (-x1, y1),
            radius=r1,
            color='k',
            fill=fill,
            alpha=alpha
        )
        b2 = plt.Circle(
            (-x2, y2),
            radius=r2,
            color='k',
            fill=fill,
            alpha=alpha
        )
        
        ax.add_patch(b1)
        ax.add_patch(b2)
        
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    ax.set_axis_off()
    
def snapshots(sys, axs, t, r1, r2, ld_params=None, cmap=plt.cm.copper, fill=True):
    
    """
    Draw a series of snapshots representing the system and multiple times.
    
    Args:
        sys (System): The system
        axs (list): A list of axes of length len(t) on which to plot the snapshots
        t (array): Time at which to draw snapshot of system
        r1 (float): Radius of the first body
        r2 (float): Radius of the second body
        ld_params (optional, float): Limb-darkening parameters for the star
        cmap (optional, matplotlib.colors.Colormap): Colormap for coloring the star according to ld_params
        fill (optional, bool): If true, fill transiting bodies, if false draw outlines only
    """
    
    if len(axs) != len(t):
        raise ValueError('One axis object is required for each time point.')
    
    for ax, ti in zip(axs, t):
        draw(sys, ax, ti, r1, r2, ld_params=ld_params, cmap=cmap, fill=fill)
        
def animate(sys, fig, t, r1, r2, duration=5, ld_params=None, cmap=plt.cm.copper):
    
    """
    Make an animation of the system.
    
    Args:
        sys (System): The system
        fig (matplotlib.pyplot.figure): Figure object
        t (array): Array of times representing each frame of the animation
        r1 (float): Radius of the first body
        r2 (float): Radius of the second body
        ld_params (optional, float): Limb-darkening parameters for the star
        cmap (optional, matplotlib.colors.Colormap): Colormap for coloring the star according to ld_params
        
    Returns:
        matplotlib.animation.FuncAnimation: The animation.
    """
        
    t = np.ascontiguousarray(t)
    ax = fig.gca()
            
    x1, y1, z1, x2, y2, z2 = sys.kep.coords(t, sys.pdict)
        
    if ld_params is None:
        star = plt.Circle((0, 0), radius=1, color=cmap(1.0), fill=True)
        ax.add_patch(star)
        
    else:
        u1, u2 = ld_params
            
        d = 0.01
        x = np.arange(-1.0, 1.0, d)
        y = np.arange(-1.0, 1.0, d)
        x, y = np.meshgrid(x, y)
        shape = np.shape(x)
        x = x.flatten()
        y = y.flatten()
        z = np.zeros(len(x))
        with np.errstate(invalid='ignore'):
            for i, (x, y) in enumerate(zip(x, y)):
                z[i] = 1 - u1 * (1 - np.sqrt(1 - x**2 - y**2)) - u2 * (1 - np.sqrt(1 - x**2 - y**2))**2
        z = z.reshape(shape)
            
        cmap = cmap
        cmap.set_bad('white')

        edge = plt.Circle(
            (0.005, -0.0), 
            radius=1.0, 
            color='w',
            fill=False, 
            linewidth=4
        )
        
        ax.imshow(z, interpolation='bilinear', extent=(-1, 1, -1, 1), cmap=cmap)
        ax.add_patch(edge)
        
    interval = int(duration * 1000 / len(t))
        
    p_patch = ax.add_patch(plt.Circle((0, 0), 0, color='k'))
    m_patch = ax.add_patch(plt.Circle((0, 0), 0, color='k'))
        
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
        
    ax.set_axis_off()
        
    def init():

        p_patch.set_center((0, 0))
        p_patch.set_radius(r1)
        m_patch.set_center((0, 0))
        m_patch.set_radius(r2)
        return p_patch, m_patch
        
    def update(i):

        p_patch.set_center((-x1[i], y1[i]))
        m_patch.set_center((-x2[i], y2[i]))
        return p_patch, m_patch
        
    return animation.FuncAnimation(
        fig, 
        update, 
        frames=np.arange(len(t)), 
        init_func=init, 
        blit=True, 
        interval=interval
    )


import numpy as np
import ctypes
from ctypes import byref
import os
import fnmatch

path, _ = os.path.split(__file__)
libname = fnmatch.filter(os.listdir(path), 'photlib*.so')[0]
clib = ctypes.CDLL(path + "/" + libname)

def flux(c1, c2, rp, rm, bp, bpm, cth, sth):
    
    j = len(bp)
    
    bp, bpm, cth, sth = map(
        lambda a: byref((ctypes.c_double * j).from_buffer(a)), 
        (bp, bpm, cth, sth, )
    )
    
    rp, rm, c1, c2 = map(
        lambda a: byref(ctypes.c_double(a)), 
        (rp, rm, c1, c2)
    )
    
    lc = ((ctypes.c_double * 8) * j).from_buffer(np.zeros((8, j)))
    j = byref(ctypes.c_int(j))
    clib.flux.restype = None
    clib.flux(c1, c2, rp, rm, bp, bpm, cth, sth, lc, j)
    return np.array(lc)

def flux_ng(c1, c2, rp, rm, bp, bpm, cth, sth):
    
    j = len(bp)
    
    bp, bpm, cth, sth = map(
        lambda a: byref((ctypes.c_double * j).from_buffer(a)), 
        (bp, bpm, cth, sth, )
    )
    
    rp, rm, c1, c2 = map(
        lambda a: byref(ctypes.c_double(a)), 
        (rp, rm, c1, c2)
    )
    
    lc = (ctypes.c_double * j).from_buffer(np.zeros(j))
    j = byref(ctypes.c_int(j))
    clib.flux_ng.restype = None
    clib.flux_ng(c1, c2, rp, rm, bp, bpm, cth, sth, lc, j)
    return np.array(lc)


import numpy as np
import ctypes
from ctypes import byref

import os
import fnmatch

path, _ = os.path.split(__file__)
libname = fnmatch.filter(os.listdir(path), 'keplib*.so')[0]
clib = ctypes.CDLL(path + "/" + libname)

__args__ = [
    'a1', 
    't1', 
    'e1', 
    'p1', 
    'w1', 
    'i1',      
    'a2', 
    't2', 
    'e2', 
    'p2', 
    'o2', 
    'w2', 
    'i2', 
    'm2'
]

def validate_elements(argdict, arglist):
    
    if set(argdict.keys()) == set(arglist):
        
        argdict = {
            k: byref(ctypes.c_double(v)) 
            for k, v in argdict.items()
        }
        idm = {v: i for i, v in enumerate(arglist)}
        args = tuple(
            dict(
                sorted(
                    argdict.items(), 
                    key=lambda p: idm[p[0]])
            ).values()
        )
        return args
        
    else:
        raise ValueError(
            "required parameters are: ", 
            arglist
        )
        
def RPP(M, ecc):
    
    j = len(M)
    M = byref((ctypes.c_double * j).from_buffer(M))
    
    for e in ecc:
        e = byref(ctypes.c_double(e))
        cosf = (ctypes.c_double * j).from_buffer(np.zeros(j))
        sinf = (ctypes.c_double * j).from_buffer(np.zeros(j))
        RPP = clib.kepler_solve_RPP(M, e, cosf, sinf, byref(ctypes.c_int(j)))
        return np.array(cosf), np.array(sinf)
    
def newton(M, ecc):
    
    j = len(M)
    M = byref((ctypes.c_double * j).from_buffer(M))
    
    for e in ecc:
        e = byref(ctypes.c_double(e))
        cosf = (ctypes.c_double * j).from_buffer(np.zeros(j))
        sinf = (ctypes.c_double * j).from_buffer(np.zeros(j))
        newton = clib.kepler_solve(M, e, cosf, sinf, byref(ctypes.c_int(j)))
        return np.array(cosf), np.array(sinf)

class Kepler:
    
    def __init__(self, args, lib):
        
        self.lib = lib
        self.args = args

    def impacts(self, t, argdict):
    
        j = len(t)
        bp, bpm, theta = tuple(
            (ctypes.c_double * j).from_buffer(np.zeros(j)) 
            for a in range(3)
        )
        t = byref((ctypes.c_double * j).from_buffer(t))
    
        args = validate_elements(argdict, self.args)
        self.lib.impacts.restype = None
        self.lib.impacts(
            t, 
            *args, 
            byref(ctypes.c_int(j)), bp, bpm, theta
        )
    
        return map(np.array, (bp, bpm, theta))
    
    def grad_impacts(self, t, argdict):
    
        j = len(t)
        bp, bpm, theta = tuple(
            (ctypes.c_double * j).from_buffer(np.zeros(j)) 
            for a in range(3)
        )
        dbp, dbpm, dtheta = tuple(
            (
                (ctypes.c_double * j) * len(self.args)
            ).from_buffer(np.zeros((j, len(self.args)))) 
            for a in range(3)
        )
        t = byref((ctypes.c_double * j).from_buffer(t))
    
        args = validate_elements(argdict, self.args)
        self.lib.grad_impacts.restype = None
        self.lib.grad_impacts(
            t, 
            *args, 
            byref(ctypes.c_int(j)), 
            bp, 
            bpm, 
            theta, 
            dbp, 
            dbpm, 
            dtheta
        )
        
        return map(
            np.array, 
            (bp, bpm, theta, dbp, dbpm, dtheta)
        )

    def coords(self, t, argdict):
    
        self.lib.coords.restype = None
        j = len(t)
    
        xp, yp, zp, xm, ym, zm = tuple(
            (ctypes.c_double * j).from_buffer(np.zeros(j)) 
            for a in range(6)
        )
        t = byref((ctypes.c_double * len(t)).from_buffer(t))
    
        args = validate_elements(argdict, self.args)
    
        self.lib.coords(
            t, 
            *args, 
            byref(ctypes.c_int(j)), 
            xp, yp, zp, 
            xm, ym, zm
        )
    
        return map(
            np.array, 
            (xp, yp, zp, xm, ym, zm)
        )



# from setuptools import setup, Extension
from numpy.distutils.core import setup, Extension

import pathlib
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

keplib = Extension(name = 'gefera.keplib',
                   sources = ['fortran/kep.f90'],
                   extra_f90_compile_args = [
                       '-Ofast', 
                       '-unroll=12',
                       '-fmax-stack-var-size=100000'   
                   ]
                  )

hrchlib = Extension(name = 'gefera.hrchlib', 
                     sources = ['fortran/kep.f90', 
                               'fortran/hrch.f90',
                              ],
                     extra_f90_compile_args = [
                         '-Ofast', 
                         '-unroll=12',
                         '-fmax-stack-var-size=100000'
                     ]
                    )
conflib = Extension(name = 'gefera.conflib', 
                    sources = ['fortran/kep.f90', 
                               'fortran/conf.f90'
                              ],
                    extra_f90_compile_args = [
                        '-Ofast', 
                        '-unroll=12',
                        '-fmax-stack-var-size=100000'
                    ]
                   )
photlib = Extension(name = 'gefera.photlib', 
                    sources = ['fortran/ellip.f90', 
                               'fortran/phot_nograd.f90',
                               'fortran/phot.f90'
                              ],
                    extra_f90_compile_args = [
                        '-Ofast', 
                        '-unroll=12',
                        '-fmax-stack-var-size=100000'
                    ]
                   )

setup(name='gefera', 
      version='0.1',
      description='Light curves for mutual transits of limb-darkened stars',
      long_description=README,
      long_description_content_type="text/markdown",
      url='http://github.com/tagordon/gefera',
      author='Tyler Gordon',
      author_email='tagordon@uw.edu', 
      license='MIT',
      packages=['gefera'],
      install_requires=['numpy',
                        'scipy',
                        'astropy'
                       ],
      zip_safe=True,
      ext_modules = [hrchlib, 
                     conflib, 
                     photlib,
                     keplib
                    ]
      )



import numpy as np
from copy import copy
import gefera as gf
import ctypes
from ctypes import byref

impacts = gf.kep.Kepler.impacts
grad_impacts = gf.kep.Kepler.grad_impacts
RPP = gf.kep.RPP
newton = gf.kep.newton

uniform_draw = lambda l, h: np.random.rand() * (h - l) + l

def random_args_conf(dictionary=True):
    
    keys = [
        'a1', 
        't1', 
        'e1', 
        'p1', 
        'w1', 
        'i1',      
        'a2', 
        't2', 
        'e2', 
        'p2', 
        'o2', 
        'w2', 
        'i2' 
    ]
    
    a1 = uniform_draw(1, 500)
    t1 = uniform_draw(-10, 10)
    e1 = uniform_draw(0, 0.999)
    p1 = uniform_draw(1, 100)
    w1 = uniform_draw(0, np.pi*2)
    i1 = uniform_draw(0, np.pi/2)
    
    a2 = uniform_draw(1, 500)
    t2 = uniform_draw(-10, 10)
    e2 = uniform_draw(0, 0.999)
    p2 = uniform_draw(1, 100)
    o2 = uniform_draw(0, np.pi*2)
    w2 = uniform_draw(0, np.pi*2)
    i2 = uniform_draw(0, np.pi/2)
    
    if dictionary:
        argdict = {
            k: v for k, v in zip(
                keys, 
                (a1, t1, e1, p1, w1, i1, a2, t2, e2, p2, o2, w2, i2)
            )
        }
    
        return argdict
    
    else:
        return a1, t1, e1, p1, w1, i1, a2, t2, e2, p2, o2, w2, i2

def random_args_hrch(dictionary=True):
    
    keys = [
        'a1', 
        't1', 
        'e1', 
        'p1', 
        'w1', 
        'i1',      
        'a2', 
        't2', 
        'e2', 
        'p2', 
        'o2', 
        'w2', 
        'i2', 
        'm2'
    ]
    
    a1 = uniform_draw(1, 500)
    t1 = uniform_draw(-10, 10)
    e1 = uniform_draw(0, 0.999)
    p1 = uniform_draw(10, 100)
    w1 = uniform_draw(0, np.pi*2)
    i1 = uniform_draw(0, np.pi/2)
    
    a2 = uniform_draw(0.1, 20)
    t2 = uniform_draw(-10, 10)
    e2 = uniform_draw(0, 0.999)
    p2 = uniform_draw(1, 10)
    o2 = uniform_draw(0, np.pi*2)
    w2 = uniform_draw(0, np.pi*2)
    i2 = uniform_draw(0, np.pi/2)
    m2 = uniform_draw(0, 1)
    
    if dictionary:
        argdict = {
            k: v for k, v in zip(
                keys, 
                (a1, t1, e1, p1, w1, i1, a2, t2, e2, p2, o2, w2, i2, m2)
            )
        }
    
        return argdict
    
    else:
        return a1, t1, e1, p1, w1, i1, a2, t2, e2, p2, o2, w2, i2, m2
    
def flux_numerical(u1, u2, rp, rm, bp, bpm, theta, nn):
    
    xp, yp = -bp, 0.0
    xm, ym = -bp + bpm * np.cos(theta), bpm * np.sin(theta)
    
    ld = lambda p: [
        1 - u1 * (1 - np.sqrt(1 - p[0]**2 - p[1]**2)) 
        - u2 * (1 - np.sqrt(1 - p[0]**2 - p[1]**2))**2 
        for p in p
    ]
    ld1 = 1 - u1 - 2 * u2
    ld2 = u1 + 2 * u2
    ld3 = u2
    
    norm = ld1 * np.pi + ld2 * 2 * np.pi / 3 + ld3 * np.pi / 2
    rp2 = rp * rp
    rm2 = rm * rm
    bpm = np.sqrt((xp - xm)**2 + (yp - ym)**2)
    
    res = nn ** 2

    if bpm < (rp + rm):
        
        if bpm + rm > rp:
            scale = 2 * (bpm + rm) + rm / 10
        else: 
            scale = 2 * rp + rm / 10
        
        n = int(np.sqrt(res * (4 * rp) ** 2))
        rat = np.sqrt(3) / 2
        nx = int(np.sqrt(n / rat))
        ny = n // nx
        xv, yv = np.meshgrid(np.arange(nx), np.arange(ny))
        xv = xv * rat
        xv[::2, :] += rat / 2
        xv = xv / len(xv) * scale + xp - scale / 2
        yv = yv / len(yv) * scale + yp - scale / 2
        points = np.concatenate((xv.flatten()[:, None], yv.flatten()[:, None]), axis=1)
    else:
        lxp = xp - rp
        uxp = xp + rp
        lxm = xm - rm
        uxm = xm + rm
    
        lyp = yp - rp
        uyp = yp + rp
        lym = ym - rm
        uym = ym + rm
    
        nump = int(np.sqrt(res * (2 * rp) ** 2))
        rat = np.sqrt(3) / 2
        nxp = int(np.sqrt(nump / rat))
        nyp = nump // nxp
        xvp, yvp = np.meshgrid(np.arange(nxp), np.arange(nyp))
        xvp = xvp * rat
        xvp[::2, :] += rat / 2
        xvp = xvp / len(xvp) * 2 * rp + xp - rp
        yvp = yvp / len(yvp) * 2 * rp + yp - rp
        pointsp = np.concatenate((xvp.flatten()[:, None], yvp.flatten()[:, None]), axis=1)
        
        numm = int(np.sqrt(res * (2 * rm) ** 2))
        rat = np.sqrt(3) / 2
        nxm = int(np.sqrt(numm / rat))
        nym = numm // nxm
        xvm, yvm = np.meshgrid(np.arange(nxm), np.arange(nym))
        xvm = xvm * rat
        xvm[::2, :] += rat / 2
        xvm = xvm / len(xvm) * 2 * rm + xm - rm
        yvm = yvm / len(yvm) * 2 * rm + ym - rm
        pointsm = np.concatenate((xvm.flatten()[:, None], yvm.flatten()[:, None]), axis=1)
        points = np.concatenate((pointsp, pointsm))
    
    dx = points[1, 0] - points[0, 0]
    dd =  dx * dx * 2 / np.sqrt(3)
    dp2 = (points[:, 0] - xp)**2 + (points[:, 1] - yp)**2
    dm2 = (points[:, 0] - xm)**2 + (points[:, 1] - ym)**2
    ds2 = points[:, 0]**2 + points[:, 1]**2
    
    occulted = ((dp2 < rp2) | (dm2 < rm2)) & (ds2 < 1)
    return - np.sum(ld(points[occulted])) * dd / norm

def change_coords(xp, yp, xm, ym):
    
    bm2 = xm**2 + ym**2
    bm = np.sqrt(bm2)
    
    bp2 = xp**2 + yp**2
    bp = np.sqrt(bp2)
    
    bpm2 = (xm - xp)**2 + (ym - yp)**2
    bpm = np.sqrt(bpm2)

    a = bp 
    b = bpm
    c = bm
        
    if a > b:
        tmp = b
        b = a
        a = tmp
    if b > c:
        mu = c - (a - b)
    else:
        mu = b - (a - c)
    
    if bpm == 0 or (a - b + c) < 1e-15:
        theta = 0.0
    elif (a - c + b) < 1e-15:
        theta = np.pi
    else:
        theta = 2 * np.arctan(np.sqrt(((a - b) + c) * mu / ((a + (b + c)) * ((a - c) + b))))
    return bp, bpm, theta